# -*- coding: utf-8 -*-
module Enumerable
  def to_list(klass = List, *args, &block)
    klass.new!(self, *args, &block)
  end
end
def List(a=nil,&b)
  b ?  List.new!(a) {b} : List.new!(a)
end
def Set(a,&b)
  b ?  Set.new!(a) {b} :  Set.new!(a)
end
def Sequence(a)
  Sequence.new(a)
end
def Sigma(a,&block)
  if a.kind_of?(Sigma)
    a 
  elsif a.kind_of?(Quotient)
    Sigma.new!(a,block)
  elsif a.kind_of?(Natural) or a.kind_of?(Counting) or a.kind_of?(Zahlen) or a.kind_of?(Fraction) or a.kind_of?(Decimal)
  elsif
    Sigma.new!(a.to_Q,block)
  elsif a.kind_of?(NaughtClass) or a.kind_of?(EmptyListClass) or a.kind_of?(NilClass)
    a unless b
    new(Quotinet(0,1),block)
  elsif a.kind_of?(NANClass) or a.kind_of?(InfinityClass)
    a
  end
end
class List
  def List.new!(a=nil, &b)
    b ?  new(a) {b} : new(a)
  end
  private_class_method :new
  include Enumerable
  def self.[](*ary)
    new(ary)
  end
  def initialize(a, &block)
    @hash ||= Hash.new
    a.nil? and return
    if block
      with_enum(a) {|o| add(block[o]) }
    else
      merge(a)
    end
  end
  private :initialize
  def with_enum(enum, &block)
    if enum.respond_to?(:each_entry)
      enum.each_entry(&block)
    elsif enum.respond_to?(:each)
      enum.each(&block)
    else
      raise ArgumentError, "value must be enumerable"
    end
  end
  private :with_enum
  def initialize_copy(orig)
    @hash = orig.instance_variable_get(:@hash).dup
  end
  def freeze
    super
    @hash.freeze
    self
  end
  def taint
    super
    @hash.taint
    self
  end
  def untaint
    super
    @hash.untaint
    self
  end
  def size
    @hash.size
  end
  alias length size
  def empty?
    @hash.empty?
  end
  def clear
    @hash.clear
    self
  end
  def replace(enum)
    if enum.instance_of?(self.class)
      @hash.replace(enum.instance_variable_get(:@hash))
    else
      clear
      merge(enum)
    end
    self
  end
  def to_a
    @hash.keys
  end
  def flatten_merge(set, seen = List.new!)
    set.each { |e|
      if e.is_a?(List)
        if seen.include?(e_id = e.object_id)
          raise ArgumentError, "tried to flatten recursive List"
        end
        seen.add(e_id)
        flatten_merge(e, seen)
        seen.delete(e_id)
      else
        add(e)
      end
    }
    self
  end
  protected :flatten_merge
  def flatten
    self.class.new!.flatten_merge(self)
  end
  def flatten!
    detect {|e| e.is_a?(List) }?  replace(flatten()) : nil
  end
  def member?(a)
    @hash.include?(a)
  end
  alias include? member?
  def contianing_list?(li)
    li.is_a?(List) or raise ArgumentError, "value must be a List"
    return false if size < li.size
    li.all? {|a| include?(a) }
  end
  alias superset? contianing_list?
  alias >= contianing_list?
  def strict_contianing_list?(li)
    li.is_a?(List) or raise ArgumentError, "value must be a List"
    return false if size <= li.size
    li.all? {|a| include?(a) }
  end
  alias proper_superset? strict_contianing_list?
  alias > strict_contianing_list?
  def contianed_list?(li)
    li.is_a?(List) or raise ArgumentError, "value must be a List"
    return false if li.size < size
    all? {|a| li.include?(a) }
  end
  alias subset? contianed_list?
  alias <= contianed_list?
  def strictly_contianed_list?(li)
    li.is_a?(List) or raise ArgumentError, "value must be a set"
    return false if li.size <= size
    all? {|a| li.include?(a) }
  end
  alias proper_subset? strictly_contianed_list?
  alias <  strictly_contianed_list?
  def each(&block)
    block or return enum_for(__method__)
    @hash.each_key(&block)
    self
  end
  def add(a)
    @hash[a] = true
    self
  end
  alias << add
  def add?(a)
   include?(a) ?      nil :      add(a)
  end
  def delete(a)
    @hash.delete(a)
    self
  end
  def delete?(a)
    include?(a) ?      delete(a) :      nil
  end
  def delete_if
    block_given? or return enum_for(__method__)
    to_a.each {|a| @hash.delete(a) if yield(a) }
    self
  end
  def keep_if
    block_given? or return enum_for(__method__)
    to_a.each {|a| @hash.delete(a) unless yield(a) }
    self
  end
  def collect!
    block_given? or return enum_for(__method__)
    li = self.class.new!
    each {|a| li << yield(a) }
    replace(li)
  end
  alias map! collect!
  def reject!(&block)
    block or return enum_for(__method__)
    n = size
    delete_if(&block)
    size == n ? nil : self
  end
  def select!(&block)
    block or return enum_for(__method__)
    n = size
    keep_if(&block)
    size == n ? nil : self
  end
  def merge(enum)
    enum.instance_of?(self.class) ?  @hash.update(enum.instance_variable_get(:@hash)) :       with_enum(enum) { |o| add(o) }
    self
  end
  def subtract(enum)
    with_enum(enum) {|a| delete(a) }
    self
  end
  def union(a)
    dup.merge(a)
  end
  alias + union 
  alias | union 
  def difference(a)
    dup.subtract(a)
  end
  alias - difference
  def intersection(a)
    n = self.class.new!
    with_enum(a) {|b| n.add(b) if include?(b) }
    n
  end
  alias & intersection
  def symetric_difference(a)
    n = List.new!(a)
    each {|b| if n.include?(b) then n.delete(b) else n.add(b) end }
    n
  end
  alias ^ symetric_difference
  alias sym_diff symetric_difference
  def ==(other)
    if self.equal?(other)
      true
    elsif other.instance_of?(self.class)
      @hash == other.instance_variable_get(:@hash)
    elsif other.is_a?(List) && self.size == other.size
      other.all? {|a| @hash.include?(a) }
    else
      false
    end
  end
  def hash
    @hash.hash
  end
  def eql?(a)
    return false unless a.is_a?(List)
    @hash.eql?(a.instance_variable_get(:@hash))
  end
  def cartisian(a)
    # E.g.
    # {1, 2} × {red, white} = {(1, red), (1, white), (2, red), (2, white)}.
    # {1, 2, green} × {red, white, green} = {(1, red), (1, white), (1, green), (2, red), (2, white), (2, green), (green, red), (green, white), (green, green)}.
    # {1, 2} × {1, 2} = {(1, 1), (1, 2), (2, 1), (2, 2)}.
    #Some basic properties of cartesian products:
    #A × ∅ = ∅.
    #A × (B ∪ C) = (A × B) ∪ (A × C).
    #(A ∪ B) × C = (A × C) ∪ (B × C).
    #Let A and B be finite sets. Then
    #| A × B | = | B × A | = | A | × | B |.
  end
  alias * cartisian
  alias x cartisian
  def categorise_list
  # e.g.:
  #   files = List.new(Dir.glob("*.rb"))
  #   ret = files.classify { |f| File.mtime(f).year }
  #   p ret    # => {2010=>#<List: {"a.rb", "b.rb"}>,2011=>#<List: {"c.rb", "d.rb", "e.rb"}>,2012=>#<List: {"f.rb"}>}
    block_given? or return enum_for(__method__)
    h = {}
    each { |i|
      x = yield(i)
      (h[x] ||= self.class.new!).add(i)
    }
    h
  end
  alias classify  categorise_list
  def contain_by_rule(&rule)
    rule or return enum_for(__method__)
    if rule.arity == 2
      require 'tsort'
      class << dig = {}         # :nodoc:
        include TSort
        alias tsort_each_node each_key
        def tsort_each_child(node, &block)
          fetch(node).each(&block)
        end
      end
      each { |u|
        dig[u] = a = []
        each{ |v| rule.call(u, v) and a << v }
      }
      li = List.new!()
      dig.each_strongly_connected_component { |css|
        li.add(self.class.new!(css))
      }
      li
    else
      List.new!(classify(&rule).values)
    end
    alias contain_by contain_by_rule
    alias divide contain_by_rule
  end
  def inspect
    ids = (Thread.current[:__inspect_key__] ||= [])
    return sprintf('#<%s:: {...}>', self.class.name) if ids.include?(object_id)
    begin
      ids << object_id
      return sprintf('#<%s:: {%s}>', self.class, to_a.inspect[1..-2])
    ensure
      ids.pop
    end
  end
  def pretty_print(pp)
    pp.text sprintf('#<%s:: {', self.class.name)
    pp.nest(1) {
      pp.seplist(self) { |o|
        pp.pp o
      }
    }
    pp.text "}>"
  end
  alias pp pretty_print
  def pretty_print_cycle(pp)
    pp.text sprintf('#<%s:: {%s}>', self.class.name, empty? ? '' : '...')
  end
  alias pp_cycle pretty_print_cycle
end
class Set < List
  @@setup = false
  def Set.new!(*args)
    new(args)
  end
  private_class_method :new
  private :initialize
  def initialize(*args, &block)
    initialize(*args, &block)
  end
end
class Sigma<Numeric
  def Sigma.new!(a,b)
    if a.kind_of?(Sigma)
      a 
    elsif a.kind_of?(Quotient)
      new(a,b)
    elsif a.kind_of?(Natural) or a.kind_of?(Counting) or a.kind_of?(Zahlen) or a.kind_of?(Fraction) or a.kind_of?(Decimal)
    elsif
      new(a.to_Q,b)
    elsif a.kind_of?(NaughtClass) or a.kind_of?(EmptyListClass) or a.kind_of?(NilClass)
      a unless b
      new(Quotinet(0,1),b) if b
    elsif a.kind_of?(NANClass) or a.kind_of?(InfinityClass)
      a
    end
  end
  private_class_method :new
  @@epsilon = Quotient(1,100000000000)
  @@upsilon = "a very large number ceiling to signal change to scientific format"
  def Sigma::epsilon(e)
#on epsilon change flag for dirty and recalculate
  end
  def initialize(rat,irrat,abs=1)
    @absolute = abs
    @rat = rat
    @irrat = irrat
    @k = 1
    @k0 = 0
    @update_proc = proc{ ret = Quotient(0,1) ; (@k0..@k).to_a.each{|n| ret += @irrat.call(n)}; ret}
  end
  attr_accessor :k, :k0
  def convergents
    []
  end
  def irrational
    true ? false : true
  end
  def rational?
    true ? true : false
  end
  def to_a
    convergents
  end
  def to_s
    e = ""
    e = "-" if @absolute == -1
    # @update_proc
  end
  def updateApprox(&block)
    instance_exec(&block)
  end
  def updateApprox1
    @rat += @update_proc.call
    #    instance_eval(@update_eval.call)
  end
  def next
    t = @k
    @k0 = t
    @k += 1
    @rat += @update_proc.call
    t = @k
    @k0 = t
    @k += 1
  end
  private :initialize
  attr_accessor :update_proc
  def to_Q
    @rat
  end
  def to_Dec
    @rat.to_Dec
  end
end
#class List
#end
class Sequence
  def initialize
    @index0 = 0
    @indexn = 50
    @func = proc{|k|
      @a_n_minus1 = @a_n
      @a_n *= (Quotient(1) + Quotient(Zahlen(1),Natural(k+1)))
    }
    @a_n_minus1 = Quotient(1)
    @a_n = Quotient(1)
    @sequence = List.new!#([])
    (@index0..@indexn).to_a.each{|n|
      @sequence << @func.call(n)}
  end
  def to_h
  end
  def to_cols
    result = [[],[]]
    (@index0..@indexn).to_a.each{|n|
      result.at(0)<< n
      result.at(1)<< @func.call(n)}
  end
  def to_Set
    Set(@sequence)
  end
  def to_Series
    a = Sigma(@sequence.inject(:+).to_Q) {@func}
    a.k0 = @index0
    a.k = @indexn
    a
  end
end
